/*
 * Copyright 2006 University Corporation for Advanced Internet Development, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package edu.internet2.middleware.shibboleth.idp.authn.provider;

import java.io.IOException;
import java.security.Principal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.ServletContext;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.security.auth.login.LoginException;

import org.opensaml.xml.util.DatatypeHelper;
import org.opensaml.xml.util.Pair;
import org.opensaml.saml2.metadata.provider.MetadataProvider;
import org.opensaml.saml2.metadata.provider.MetadataProviderException;
import org.opensaml.util.storage.StorageService;

import edu.internet2.middleware.shibboleth.common.attribute.BaseAttribute;
import edu.internet2.middleware.shibboleth.common.attribute.provider.SAML2AttributeAuthority;
import edu.internet2.middleware.shibboleth.common.attribute.provider.ShibbolethSAML2AttributeAuthority;
import edu.internet2.middleware.shibboleth.common.attribute.AttributeRequestException;
import edu.internet2.middleware.shibboleth.common.attribute.resolver.AttributeResolver;

import edu.internet2.middleware.shibboleth.common.profile.provider.BaseSAMLProfileRequestContext;
import edu.internet2.middleware.shibboleth.common.relyingparty.RelyingPartyConfiguration;
import edu.internet2.middleware.shibboleth.common.relyingparty.RelyingPartyConfigurationManager;
import edu.internet2.middleware.shibboleth.common.relyingparty.provider.SAMLMDRelyingPartyConfigurationManager;


import edu.internet2.middleware.shibboleth.idp.authn.AuthenticationEngine;
import edu.internet2.middleware.shibboleth.idp.authn.AuthenticationException;
import edu.internet2.middleware.shibboleth.idp.authn.LoginHandler;
import edu.internet2.middleware.shibboleth.idp.authn.UsernamePrincipal;
import edu.internet2.middleware.shibboleth.idp.authn.LoginContext;
import edu.internet2.middleware.shibboleth.idp.authn.LoginContextEntry;
import edu.internet2.middleware.shibboleth.idp.authn.PassiveAuthenticationException;
import edu.internet2.middleware.shibboleth.idp.util.HttpServletHelper;

import edu.internet2.middleware.shibboleth.idp.authn.provider.UsernamePasswordLoginServlet;

public class AuthnContextLoginServlet extends UsernamePasswordLoginServlet {
// public class AuthnContextLoginServlet extends HttpServlet {

     /** Serial version UID. */
    private static final long serialVersionUID = -572723856392956990L;

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AuthnContextLoginServlet.class);

    /* silver uris */
    public static final String SILVER_AUTHN_CLASSREF = "http://incommonfederation.org/assurance/silver";
    public static final String SILVER_TEST_AUTHN_CLASSREF = "http://incommonfederation.org/assurance/silver-test";


    // these are same as for the super
    /** Parameter name to indicate login failure. */
    private final String failureParam = "loginFailed";
    
    /** HTTP request parameter containing the user name. */
    private final String usernameAttribute = "j_username";

    /** HTTP request parameter containing the user's password. */
    private final String passwordAttribute = "j_password";

    /* Servlet Context */
    private static ServletContext context;
    ShibbolethSAML2AttributeAuthority attributeAuthority;
    AttributeResolver attributeResolver;
    SAMLMDRelyingPartyConfigurationManager relyingPartyConfigurationManager;
    private static StorageService<String, LoginContextEntry> storageService;
    private boolean initialized = false;

    // will be an init param
    Map<String, String> authnMethods;

    /** {@inheritDoc} */
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        log.debug("AuthnContextLoginServlet init");
        
        // save all the uri-looking params
        authnMethods = new HashMap();
        Enumeration<String> params = getInitParameterNames();
        while (params.hasMoreElements()) {
            String k = params.nextElement();
            String v = getInitParameter(k);
            if (k.matches("^[a-z]+:.*")) authnMethods.put(k, v);
        }
          
        // if (getInitParameter(loginPageInitParam) != null) {
        //     loginPage = getInitParameter(loginPageInitParam);
        // }

        context = config.getServletContext();
        attributeAuthority = (ShibbolethSAML2AttributeAuthority) HttpServletHelper.getSAML2AttributeAuthority(context);
        attributeResolver = HttpServletHelper.getAttributeResolver(context);
        relyingPartyConfigurationManager = (SAMLMDRelyingPartyConfigurationManager) HttpServletHelper.getRelyingPartyConfirmationManager(context);
        storageService = (StorageService<String, LoginContextEntry>) HttpServletHelper.getStorageService(context);

        if (attributeAuthority!=null && attributeResolver!=null && relyingPartyConfigurationManager!=null && storageService!=null) initialized = true;
        else log.error("AuthnContextLoginServlet init failed! authncontext processing disabled.");
    }


    /** {@inheritDoc} */
    protected void service(HttpServletRequest request, HttpServletResponse response) throws ServletException,
            IOException {
        String principalName = request.getRemoteUser();

        // Is this a remote user login
        if (principalName!=null) {
           log.debug("Remote user identified as {} ... back to authentication engine", principalName);
           request.setAttribute(LoginHandler.PRINCIPAL_KEY, new UsernamePrincipal(principalName));
           verifyAuthn(request, principalName);
           AuthenticationEngine.returnToAuthenticationEngine(request, response);

        } else {

           // Maybe is username password

           String username = request.getParameter(usernameAttribute);
           String password = request.getParameter(passwordAttribute);

           if (username == null || password == null) {
               redirectToLoginPage(request, response);
               return;
           }
   
           try {
              authenticateUser(request, username, password);
              verifyAuthn(request, username);
              AuthenticationEngine.returnToAuthenticationEngine(request, response);
           } catch (LoginException e) {
              request.setAttribute(failureParam, "true");
              request.setAttribute(LoginHandler.AUTHENTICATION_EXCEPTION_KEY, new AuthenticationException(e));
              redirectToLoginPage(request, response);
           }
       }
    }


    /** 
     * Verify that the user is actually OK for the authn method
     * 
     * @param request servlet request
     * @param principalName remote user's id
     *
     */

    private void verifyAuthn(HttpServletRequest request, String principalName) {

        String actualAuthnMethod = DatatypeHelper.safeTrimOrNullString((String) request.getAttribute(LoginHandler.AUTHENTICATION_METHOD_KEY));
        LoginContext loginContext = HttpServletHelper.getLoginContext(storageService, context, request);
        if (loginContext==null) {
           log.error(".. setAuthnContext no login context");
           request.setAttribute(LoginHandler.AUTHENTICATION_ERROR_KEY, "authn internal error");
           return;
        }
        String authnMethod = loginContext.getAttemptedAuthnMethod();
        log.debug("aetAuthn: req=" + authnMethod + ", actual=" + actualAuthnMethod);
        if (actualAuthnMethod != null) authnMethod = actualAuthnMethod;

        if (!authnMethods.containsKey(authnMethod)) {
            log.debug("not one of ours");
            return;
        }

        if (!initialized) {
           log.error(".. setAuthnContext init failed earlier");
           request.setAttribute(LoginHandler.AUTHENTICATION_ERROR_KEY, "authn internal init error");
           return;
        }


        // get the attribute to verify this method
        String authnAttr = authnMethods.get(authnMethod);
        log.debug("looking for " + authnAttr);

        BaseSAMLProfileRequestContext requestContext = createRequestContext(loginContext, principalName, authnAttr);

        try {
            log.debug("Resolving attribute ");
            requestContext.setPrincipalAuthenticationMethod(authnMethod);  // incase a resolver script wants it
            Map<String, BaseAttribute> attributes = attributeAuthority.getAttributeResolver().resolveAttributes(requestContext);
            // requestContext.setAttributes(principalAttributes);

            Set<String> pas = attributes.keySet();
            Iterator<String> pasi = pas.iterator();
            while (pasi.hasNext()) {
               log.debug("attr: " + pasi.next());
            }
            
            // is there an authncontext attr?
            BaseAttribute ac = attributes.get(authnAttr);
            if (ac!=null && ac.getValues().size()>0) {
                 Collection<String> acv = ac.getValues();
                 Iterator<String> it = acv.iterator();
                 if (it.hasNext()) {
                    String v = it.next();
                    log.info("authnAttr = " + v);
                    // request.setAttribute(LoginHandler.AUTHENTICATION_METHOD_KEY, v);
                  }
             } else {
                 log.info("no authnAttr, no go on the authn method");
                 request.setAttribute(LoginHandler.AUTHENTICATION_ERROR_KEY, "No permission for " + authnMethod);
             }

        } catch (AttributeRequestException e) {
            log.warn("Error resolving attribute for authn.  No permission for " + authnMethod);
            request.setAttribute(LoginHandler.AUTHENTICATION_ERROR_KEY, "No permission for " + authnMethod);
        }
    
    }
   

    private BaseSAMLProfileRequestContext createRequestContext(LoginContext loginContext, String principalName, String attrId) {
        log.debug("AuthnContextLoginServlet createRequestContext");

        BaseSAMLProfileRequestContext context = new BaseSAMLProfileRequestContext();
        String rpId = loginContext.getRelyingPartyId();
        log.debug("AuthnContextLoginServlet rpid = " + rpId);
        RelyingPartyConfiguration relyingPartyConfiguration = relyingPartyConfigurationManager.getRelyingPartyConfiguration(rpId);
        MetadataProvider metadataProvider = relyingPartyConfigurationManager.getMetadataProvider();
        String idpId = relyingPartyConfiguration.getProviderId();

        Collection<String> attrs = new Vector(1);
        attrs.add(attrId);

        context.setRelyingPartyConfiguration(relyingPartyConfiguration);
        context.setInboundMessageIssuer(rpId);
        context.setOutboundMessageIssuer(idpId);
        context.setPrincipalName(principalName);
        context.setLocalEntityId(idpId);
        context.setPeerEntityId(rpId);
        context.setMetadataProvider(metadataProvider);
        context.setRequestedAttributes(attrs);

        try {
           context.setLocalEntityMetadata(metadataProvider.getEntityDescriptor(idpId));
           context.setPeerEntityMetadata(metadataProvider.getEntityDescriptor(rpId));
        } catch (MetadataProviderException e) {
           log.error("no metadata" +  e);
           return null;
        }
        return context;
    }

}
